/*
 * This file is part of FGGL.
 *
 * FGGL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * FGGL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with FGGL.
 * If not, see <https://www.gnu.org/licenses/>.
 */

//
// Created by webpigeon on 27/06/22.
//

#ifndef FGGL_MODULES_MANAGER_HPP
#define FGGL_MODULES_MANAGER_HPP

#include "fggl/debug/logging.hpp"
#include "fggl/util/graph.hpp"

#include "fggl/services/service.hpp"
#include "fggl/services/factory.hpp"
#include "fggl/services/module.hpp"

#include <format>
#include <cassert>
#include <queue>
#include <vector>
#include <map>
#include <stack>
#include <set>
#include <iostream>

namespace fggl::modules {

	/**
	 * Store and initialise modules present in the engine.
	 *
	 * This class is responsible for keeping track of which modules the library user has requested, and ensuring that
	 * their dependencies are loaded in the correct order. Once the dependency graph has been built and instances
	 * created, it is responsible for providing references to these initialised classes.
	 */
	class Manager {
		public:
			Manager() = default;

			Manager(Manager&) = delete;
			void operator=(Manager&) = delete;

			template<services::Subclass<services::Module> T>
			void use() {
				static_assert( !std::is_abstract_v<T>, "Module provided is abstract, check you have implemented all required methods." );
				assert(!m_locked);
				auto modulePtr = std::make_shared<T>();
				depend(modulePtr);
			}

			inline void depend(std::shared_ptr<services::Module>&& modulePtr) {
				assert(!m_locked);
				m_modules[modulePtr->getName()] = modulePtr;
				for ( auto& provided : modulePtr->getProvides() ) {
					m_serviceProviders[provided] = modulePtr->getName();
				}
			}

			template<services::ServiceAPI T>
			auto get() -> T * {
				assert(m_locked);
				return m_services.template getLazy<T>().ptr();
			}

			template<services::ServiceAPI T>
			auto get() const -> std::optional<T> * {
				assert(m_locked);
				return m_services.template getFromCache<T>();
			}

			auto resolve() -> bool;

			inline auto* services() {
				return &m_services;
			}

		private:
			bool m_locked = false;
			services::Generator m_services;
			std::map<services::ModuleID, std::shared_ptr<services::Module> > m_modules;

			ds::DirectedGraph<services::ModuleID> m_dependencies;
			std::map<services::ModuleID, services::ModuleID> m_serviceProviders;

			bool buildGraph();
	};

} // namespace fggl::modules

#endif //FGGL_MODULES_MANAGER_HPP
